﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using CG2.Rendering;
using CG2.Mathematics;
using CG2.Modeling;

namespace CG2.Ex07
{
    public struct Point
    {
        public Vector4 Position;
        public bool Selected;
    }

    public partial class Form1 : Form
    {
        World world = new World();
        public static Random Rnd = new Random(0);
        Point Direction;
        Point Origin;
        bool Down = false, BuildFirst = true, Trace = false;
        Ray ray;
        KdTreeAccelerator KdTreeAcceler;

        public Form1()
        {
            InitializeComponent();
            Origin.Position = new Vector4(200, 115, 0);
            Direction.Position = new Vector4(200, 280, 0);
            InitSceneSpheres();
        }

        protected override void OnPaint(PaintEventArgs e)
        {
            Graphics g = e.Graphics;

            //Info: Draw all spheres
            foreach (Model model in world.MeshModels)
            {
                Sphere s = model as Sphere;
                Pen p = new Pen(Color.FromArgb((int)s.Color.X * 255, (int)s.Color.Y * 255, (int)s.Color.Z * 255));
                p.Width = 1;
                g.DrawEllipse(p, (float)s.Min.X, (float)s.Min.Y, (float)s.Radius * 2, (float)s.Radius * 2);
            }

            //Info: Build
            if (!BuildFirst)
            {
                KdTreeAcceler = new KdTreeAccelerator();
                KdTreeAcceler.BuildStructure(world, g);
                BuildFirst = true;
                label1.Text = "Level: ";// +KdTreeAcceler.GetMaxLevel();
            }

            //Info: If the ray is no longer null, collide it with the scene
            //      If the hit model is not null, draw the hit model - intersected sphere
            ray = new Ray(Origin.Position, (Direction.Position - Origin.Position));
            if (ray != null)
            {
                Vector4 point = ray.Origin + 1 * ray.Direction;
                g.DrawLine(Pens.Black, (float)ray.Origin.X, (float)ray.Origin.Y, (float)(ray.Origin + 5 * ray.Direction).X, (float)(ray.Origin + 5 * ray.Direction).Y);
                g.FillEllipse(Brushes.Blue, (float)ray.Origin.X - 5, (float)ray.Origin.Y - 5, 10, 10); //start
                g.FillEllipse(Brushes.Green, (float)point.X - 5, (float)point.Y - 5, 10, 10); //end

                if ((KdTreeAcceler != null) && Trace)
                    KdTreeAcceler.Collide(ray, g);

                if (ray.HitModel != null)
                {
                    g.FillEllipse(Brushes.LightGreen, (float)ray.HitModel.Min.X, (float)ray.HitModel.Min.Y, (float)ray.HitModel.Radius * 2, (float)ray.HitModel.Radius * 2);
                    Pen p = new Pen(Color.FromArgb((int)ray.HitModel.Color.X * 255, (int)ray.HitModel.Color.Y * 255, (int)ray.HitModel.Color.Z * 255));
                    p.Width = 1;
                    g.DrawEllipse(p, (float)ray.HitModel.Min.X, (float)ray.HitModel.Min.Y, (float)ray.HitModel.Radius * 2, (float)ray.HitModel.Radius * 2);
                }
            }
        }

        public void InitSceneSpheres()
        {
            for (Int32 i = 0; i < 60; i++)
            {
                Sphere sphere = new Sphere(new Vector4(Rnd.Next(1, 500), Rnd.Next(1, 500), 0), 8);
                world.MeshModels.Add(sphere);
            }
        }

        private void button2_Click(object sender, EventArgs e)
        {
            BuildFirst = false;
            Invalidate();
        }

        private void button4_Click(object sender, EventArgs e)
        {
            Trace = true;
            Invalidate();
        }

        public bool Select(Vector4 p)
        {
            Vector4 d = p - Origin.Position;
            if (d.Length < 10)
            {
                Origin.Selected = true;
                Direction.Selected = false;
                return true;
            }

            d = p - Direction.Position;
            if (d.Length < 10)
            {
                Direction.Selected = true;
                Origin.Selected = false;
                return true;
            }
            return false;
        }

        protected override void OnMouseDown(MouseEventArgs e)
        {
            Vector4 point = new Vector4(e.X, e.Y, 0, 1);

            //If clicked point can be activated, then activate it and return true
            if (Select(point))
            {
                Down = true;
                Invalidate();
            }
            Invalidate();
        }

        protected override void OnMouseUp(MouseEventArgs e)
        {
            Down = false;
        }

        protected override void OnMouseMove(MouseEventArgs e)
        {
            Vector4 point = new Vector4(e.X, e.Y, 0, 1);
            if (Down)
            {
                if (Origin.Selected)
                    Origin.Position = point;
                if (Direction.Selected)
                    Direction.Position = point;
                Invalidate();
            }
        }
    }
}
